/*
 * sch_edf.c	EDF queue.
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 *
 * Authors:	Dimitry Ketov <dimitry.ketov@avalon.ru>
 */

#include <linux/config.h>
#ifdef CONFIG_MODVERSIONS
#define MODVERSIONS
#include <linux/modversions.h>
#endif
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/skbuff.h>
#include <linux/rtnetlink.h>
#include <net/ip.h>
#include <net/pkt_sched.h>

#ifdef DEBUG
#define dprintk(format, a...) printk(KERN_INFO "%s: " format, __FUNCTION__, ##a)
#else
#define dprintk(format, a...)
#endif

/* #if PSCHED_CLOCK_SOURCE == PSCHED_JIFFIES */
/* #error This version of EDF qdisc will work incorrectly (due to timer */
/* #error resolution) if PSCHED_SOURCE is set to PSCHED_JIFFIES. */
/* #endif */

#if PSCHED_CLOCK_SOURCE == PSCHED_GETTIMEOFDAY
#define PSCHED_TIME_T2US(x) ((x).tv_sec * 1000000 + (x).tv_usec)
#else
#define PSCHED_TIME_T2US(x) (x)
#endif

struct tc_edf_qopt {
	unsigned delay;
	unsigned limit;
};

struct tc_edf_copt {
	unsigned delay;
	unsigned limit;
};

struct edf_pkt_deadline {
	psched_time_t deadline;

	struct edf_pkt_deadline *next;
};

struct edf_flow {
	struct edf_flow *next;
	struct edf_flow *prev;

	u32 classid;

	unsigned long delay;
	unsigned limit;

	struct sk_buff_head q;	/* edf 'classes' do not use other qdiscs */

	struct edf_pkt_deadline *pktHead;
	struct edf_pkt_deadline *pktTail;
};

struct edf_sched_data {
	struct tcf_proto *filterList;

	struct edf_flow *flowHead;
	struct edf_flow *flowTail;
};

static struct sk_buff *edf_flow_dequeue(struct edf_flow *flow)
{
	struct edf_pkt_deadline *pkt;
	struct sk_buff *skb;

	skb = skb_dequeue(&flow->q);

	if (skb != NULL) {
		pkt = flow->pktHead;
		flow->pktHead = pkt->next;

		kfree(pkt);
	}

	return skb;
}

static int edf_flow_queue(struct edf_flow *flow, struct sk_buff *skb,
			  psched_time_t deadline)
{
	struct edf_pkt_deadline *pkt;

	pkt = kmalloc(sizeof(struct edf_pkt_deadline), GFP_ATOMIC);
	if (pkt == NULL) {
		return -ENOMEM;
	}

	pkt->next = NULL;
	pkt->deadline = deadline;

	if (flow->pktHead != NULL) {
		flow->pktTail->next = pkt;
		flow->pktTail = pkt;
	} else {
		flow->pktHead = flow->pktTail = pkt;
	}

	skb_queue_tail(&flow->q, skb);

	return 0;
}

static void edf_flow_reset(struct edf_flow *flow)
{
	struct edf_pkt_deadline *pkt;

	skb_queue_purge(&flow->q);

	while ((pkt = flow->pktHead) != NULL) {
		flow->pktHead = pkt->next;
		kfree(pkt);
	}

	flow->pktTail = flow->pktHead = NULL;
}

static void edf_flow_init(struct edf_flow *flow, u32 classid)
{
	skb_queue_head_init(&flow->q);

	flow->next = NULL;
	flow->prev = NULL;
	flow->pktHead = NULL;
	flow->pktTail = NULL;

	flow->delay = 0;
	flow->classid = classid;
}

static struct edf_flow *edf_flow_find(struct edf_flow *flow, u32 classid)
{
	while (flow != NULL && flow->classid != classid) {
		flow = flow->next;
	}

	return flow;
}

static int edf_class_graft(struct Qdisc *sch, unsigned long cl,
			   struct Qdisc *new, struct Qdisc **old)
{
	return 1;
}

static struct Qdisc *edf_class_leaf(struct Qdisc *sch, unsigned long cl)
{
	return NULL;
}

static unsigned long edf_class_get(struct Qdisc *sch, u32 classid)
{
	struct edf_sched_data *q = (struct edf_sched_data *) sch->data;

	return (unsigned long) edf_flow_find(q->flowHead, classid);
}

static unsigned long edf_filter_bind(struct Qdisc *sch,
				     unsigned long parent, u32 classid)
{
	return edf_class_get(sch, classid);
}

static void edf_class_put(struct Qdisc *sch, unsigned long cl)
{
	return;
}

static void edf_filter_unbind(struct Qdisc *sch, unsigned long cl)
{
	edf_class_put(sch, cl);

	return;
}

static int edf_class_change(struct Qdisc *sch, u32 classid, u32 parent,
			    struct rtattr **tca, unsigned long *cl)
{
	struct edf_sched_data *q = (struct edf_sched_data *) sch->data;
	struct rtattr *opt = tca[TCA_OPTIONS - 1];
	struct edf_flow *flow = (struct edf_flow *) *cl;
	struct tc_edf_copt *copt;

	if (opt == NULL || RTA_PAYLOAD(opt) < sizeof(struct tc_edf_copt)) {
		return -EINVAL;
	}

	copt = RTA_DATA(opt);

	if (flow != NULL) {
		flow->delay = copt->delay;
		flow->limit = copt->limit;

		return 0;
	} else {
		flow = kmalloc(sizeof(struct edf_flow), GFP_ATOMIC);
		if (flow == NULL) {
			return -ENOMEM;
		}

		edf_flow_init(flow, classid);
		flow->delay = copt->delay;
		flow->limit = copt->limit;

		if (q->flowHead != NULL) {
			flow->prev = q->flowTail;
			q->flowTail->next = flow;
			q->flowTail = flow;
		} else {
			q->flowHead = q->flowTail = flow;
		}

		*cl = (unsigned long) flow;
	}

	return 0;
}

static int edf_class_delete(struct Qdisc *sch, unsigned long cl)
{
	struct edf_sched_data *q = (struct edf_sched_data *) sch->data;
	struct edf_flow *flow = (struct edf_flow *) cl;

	if (flow == q->flowHead) {
		return -EBUSY;
	}

	edf_flow_reset(flow);

	if (flow->next != NULL) {	/* if not the last one */
		flow->next->prev = flow->prev;
	} else {		/* the last one */

		q->flowTail = flow->prev;
	}

	if (flow->prev != NULL) {	/* if not the frist one */
		flow->prev->next = flow->next;
	} else {		/* the first one */

		q->flowHead = flow->next;
	}

	kfree(flow);

	return 0;
}

static void edf_class_walk(struct Qdisc *sch, struct qdisc_walker *walker)
{
	struct edf_sched_data *q = (struct edf_sched_data *) sch->data;
	struct edf_flow *flow;

	for (flow = q->flowHead; flow != NULL; flow = flow->next) {
		walker->fn(sch, (unsigned long) flow, walker);
		walker->count++;
	}

	return;
}

static int edf_class_dump(struct Qdisc *q, unsigned long cl,
			  struct sk_buff *skb, struct tcmsg *tcm)
{
	struct edf_flow *flow = (struct edf_flow *) cl;
	unsigned char *b = skb->tail;
	struct tc_edf_copt copt;

	copt.delay = flow->delay;
	copt.limit = flow->limit;

	RTA_PUT(skb, TCA_OPTIONS, sizeof(copt), &copt);
	tcm->tcm_parent = TC_H_ROOT;
	tcm->tcm_handle = flow->classid;
	tcm->tcm_info = 0;

	return skb->len;

      rtattr_failure:
	skb_trim(skb, b - skb->data);
	return -1;
}

static struct tcf_proto **edf_filter_chain(struct Qdisc *sch,
					   unsigned long cl)
{
	struct edf_sched_data *q = (struct edf_sched_data *) sch->data;

	return &q->filterList;
}

static int edf_drop(struct Qdisc *sch)
{
	struct edf_sched_data *q = (struct edf_sched_data *) sch->data;
	struct sk_buff *skb;

	/* 'default' flow (first flow, at the head of flow list)
	   is not real-time one, so try to drop on packet from it */
	skb = edf_flow_dequeue(q->flowHead);

	if (skb != NULL) {
		sch->stats.backlog -= skb->len;
		kfree_skb(skb);
		return 1;
	}

	return 0;
}

static void edf_reset(struct Qdisc *sch)
{
	struct edf_sched_data *q = (struct edf_sched_data *) sch->data;
	struct edf_flow *flow;

	for (flow = q->flowHead; flow != NULL; flow = flow->next) {
		edf_flow_reset(flow);
	}

	sch->stats.backlog = 0;

	return;
}

static int edf_enqueue(struct sk_buff *skb, struct Qdisc *sch)
{
	struct edf_sched_data *q = (struct edf_sched_data *) sch->data;
	struct edf_flow *flow = NULL;
	struct tcf_result result;
	psched_time_t deadline;
	int rc;

	PSCHED_GET_TIME(deadline);
	dprintk("E:NOW=%lu\n", (unsigned long) PSCHED_TIME_T2US(deadline));

	rc = tc_classify(skb, q->filterList, &result);
	dprintk("rc=%d, classid=0x%04x\n", rc, result.classid);

	if (rc >= 0) {
		flow = edf_flow_find(q->flowHead, result.classid);
	}

	if (flow == NULL) {
		flow = q->flowHead;	/* use 'default' flow */
	}

	if (flow->q.qlen <= flow->limit) {
		PSCHED_TADD(deadline, flow->delay);
		dprintk("deadline=%lu\n",
			(unsigned long) PSCHED_TIME_T2US(deadline));
		dprintk("flow=0x%p, skb=0x%p\n", flow, skb);

		if (edf_flow_queue(flow, skb, deadline) == 0) {
			sch->stats.backlog += skb->len;
			sch->stats.bytes += skb->len;
			sch->stats.packets++;
			return 0;
		}
	}
	sch->stats.drops++;
	kfree_skb(skb);

	return NET_XMIT_DROP;
}

static int edf_requeue(struct sk_buff *skb, struct Qdisc *sch)
{
	/* We've missed deadline anyway, just drop the packet :( */
	sch->stats.drops++;
	kfree_skb(skb);

	return 0;
}

static struct sk_buff *edf_dequeue(struct Qdisc *sch)
{
	struct edf_sched_data *q = (struct edf_sched_data *) sch->data;
	struct edf_flow *flow, *minflow = NULL;
	struct sk_buff *skb;
	psched_time_t mindeadline, now;

	PSCHED_GET_TIME(now);
	PSCHED_SET_PASTPERFECT(mindeadline);

	/* start from first real-time flow, skip 'default' flow */
	for (flow = q->flowHead->next; flow != NULL; flow = flow->next) {
		if (flow->pktHead != NULL) {	/* flow is not empty */
			if (minflow == NULL ||	/* not set yet - got candidate */
			    PSCHED_TLESS(flow->pktHead->deadline,
					 mindeadline)) {
				mindeadline = flow->pktHead->deadline;
				minflow = flow;
			}
		}
	}

	if (minflow == NULL) {
		minflow = q->flowHead;	/* by default try 'deafult' flow */
	}

	if ((skb = edf_flow_dequeue(minflow))) {
		sch->stats.backlog -= skb->len;
	}

	dprintk("D:NOW=%lu\n", (unsigned long) PSCHED_TIME_T2US(now));
	dprintk("minflow=0x%p, skb=0x%p\n", minflow, skb);

	return skb;
}

static int edf_init(struct Qdisc *sch, struct rtattr *opt)
{
	struct edf_sched_data *q = (struct edf_sched_data *) sch->data;
	struct tc_edf_qopt *qopt = RTA_DATA(opt);

	q->flowHead = kmalloc(sizeof(struct edf_flow), GFP_ATOMIC);
	if (q->flowHead == NULL) {
		return -ENOMEM;
	}

	edf_flow_init(q->flowHead, sch->handle);

	if (opt == NULL) {
		q->flowHead->limit = sch->dev->tx_queue_len;
		q->flowHead->delay = 0;
	} else {
		if (RTA_PAYLOAD(opt) < sizeof(struct tc_edf_qopt)) {
			return -EINVAL;
		}

		q->flowHead->limit = qopt->limit;
		q->flowHead->delay = qopt->delay;
	}

	q->flowTail = q->flowHead;
	q->filterList = NULL;

	MOD_INC_USE_COUNT;

	return 0;
}

static int edf_change(struct Qdisc *sch, struct rtattr *opt)
{
	struct edf_sched_data *q = (struct edf_sched_data *) sch->data;
	struct tc_edf_qopt *qopt = RTA_DATA(opt);

	if (opt == NULL || RTA_PAYLOAD(opt) < sizeof(struct tc_edf_qopt)) {
		return -EINVAL;
	}

	q->flowHead->limit = qopt->limit;
	q->flowHead->delay = qopt->delay;

	return 0;
}

static void edf_destroy(struct Qdisc *sch)
{
	struct edf_sched_data *q = (struct edf_sched_data *) sch->data;
	struct edf_flow *flow;
	struct tcf_proto *tp;

	edf_reset(sch);

	while ((tp = q->filterList) != NULL) {
		q->filterList = tp->next;
		tp->ops->destroy(tp);
	}

	while ((flow = q->flowHead) != NULL) {
		q->flowHead = flow->next;
		kfree(flow);
	}

	MOD_DEC_USE_COUNT;

	return;
}

static int edf_dump(struct Qdisc *sch, struct sk_buff *skb)
{
	struct edf_sched_data *q = (struct edf_sched_data *) sch->data;
	unsigned char *b = skb->tail;
	struct tc_edf_qopt opt;

	opt.limit = q->flowHead->limit;
	opt.delay = q->flowHead->delay;

	RTA_PUT(skb, TCA_OPTIONS, sizeof(opt), &opt);

	return skb->len;

      rtattr_failure:
	skb_trim(skb, b - skb->data);
	return -1;
}


static struct Qdisc_class_ops edf_class_ops = {
	graft:       edf_class_graft,
	leaf:        edf_class_leaf,
	get:         edf_class_get,
	put:         edf_class_put,
	change:      edf_class_change,
	delete:      edf_class_delete,
	walk:        edf_class_walk,
	tcf_chain:   edf_filter_chain,
	bind_tcf:    edf_filter_bind,
	unbind_tcf:  edf_filter_unbind,
	dump:        edf_class_dump,
};

struct Qdisc_ops edf_qdisc_ops = {
	next:        NULL,
	cl_ops:      &edf_class_ops,
	id:          "edf",
	priv_size:   sizeof(struct edf_sched_data),
	enqueue:     edf_enqueue,
	dequeue:     edf_dequeue,
	requeue:     edf_requeue,
	drop:        edf_drop,
	init:        edf_init,
	reset:       edf_reset,
	destroy:     edf_destroy,
	change:      edf_change,
	dump:        edf_dump,
};

static int __init edf_module_init(void)
{
	return register_qdisc(&edf_qdisc_ops);
}

static void __exit edf_module_cleanup(void)
{
	unregister_qdisc(&edf_qdisc_ops);
}

module_init(edf_module_init);
module_exit(edf_module_cleanup);

MODULE_LICENSE("GPL");
